# -*- coding: utf-8 -*-
# Time       : 2023/8/20 0:16
# Author     : QIN2DIM
# GitHub     : https://github.com/QIN2DIM
# Description:
from __future__ import annotations

import time
from abc import ABC
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, Literal
from urllib.parse import quote

from loguru import logger

from hcaptcha_challenger.onnx.modelhub import ModelHub
from hcaptcha_challenger.onnx.resnet import ResNetControl
from hcaptcha_challenger.onnx.yolo import YOLOv8

HOOK_CHALLENGE = "//iframe[contains(@src,'#frame=challenge')]"


@dataclass
class Status:
    # <success> Challenge Passed by following the expected
    CHALLENGE_SUCCESS = "success"
    # <continue> Continue the challenge
    CHALLENGE_CONTINUE = "continue"
    # <crash> Failure of the challenge as expected
    CHALLENGE_CRASH = "crash"
    # <retry> Your proxy IP may have been flagged
    CHALLENGE_RETRY = "retry"
    # <refresh> Skip the specified label as expected
    CHALLENGE_REFRESH = "refresh"
    # <backcall> (New Challenge) Types of challenges not yet scheduled
    CHALLENGE_BACKCALL = "backcall"
    # <to-X> NOT MATCH PATTERN
    CHALLENGE_TO_BINARY = "to_binary"
    CHALLENGE_TO_AREA_SELECT = "to_area_select"

    AUTH_SUCCESS = "success"
    AUTH_ERROR = "error"
    AUTH_CHALLENGE = "challenge"


@dataclass
class Skeleton(ABC):
    """hCAPTCHA challenge drive control"""

    modelhub: ModelHub
    """
    Build Skeleton with modelhub
    """

    this_dir: Path = Path(__file__).parent
    """
    Project directory of Skeleton Agents
    """

    tmp_dir: Path = this_dir.joinpath("temp_cache")
    challenge_dir = tmp_dir.joinpath("_challenge")
    """
    Runtime cache
    """

    _label = ""
    """
    Cleaned Challenge Prompt in the context
    """

    _prompt = ""
    """
    Challenge Prompt in the context
    """

    _label_alias: Dict[str, str] = field(default_factory=dict)
    """
    A collection of { prompt[s]: model_name[.onnx] }
    """

    _alias2locator: dict = field(default_factory=dict)
    """
    Store the `element locator` of challenge images {挑战图片1: locator1, ...}
    """

    _alias2url: dict = field(default_factory=dict)
    """
    Store the `download link` of the challenge image {挑战图片1: url1, ...}
    """

    _alias2path: dict = field(default_factory=dict)
    """
    Store the `directory` of challenge image {挑战图片1: "/images/挑战图片1.png", ...}
    """

    @property
    def status(self):
        return Status

    @classmethod
    def from_modelhub(cls, tmp_dir: Path | None = None, **kwargs):
        modelhub = ModelHub.from_github_repo(**kwargs)
        modelhub.parse_objects()

        dragon = cls(modelhub=modelhub, _label_alias=modelhub.label_alias)

        if tmp_dir and isinstance(tmp_dir, Path):
            dragon.tmp_dir = tmp_dir
            dragon.challenge_dir = tmp_dir.joinpath("_challenge")

        return dragon

    def _match_solution(self, select: Literal["yolo", "resnet"] = None) -> ResNetControl | YOLOv8:
        """match solution after `tactical_retreat`"""
        focus_label = self._label_alias.get(self._label, "")

        # Match YOLOv8 model
        if not focus_label or select == "yolo":
            focus_name, classes = self.modelhub.apply_ash_of_war(ash=self._label)
            session = self.modelhub.match_net(focus_name=focus_name)
            detector = YOLOv8.from_pluggable_model(session, classes)
            return detector

        # Match ResNet model
        focus_name = focus_label
        if not focus_name.endswith(".onnx"):
            focus_name = f"{focus_name}.onnx"
        net = self.modelhub.match_net(focus_name=focus_name)
        control = ResNetControl.from_pluggable_model(net)
        return control

    def switch_to_challenge_frame(self, ctx, **kwargs):
        raise NotImplementedError

    def get_label(self, ctx, **kwargs):
        """Obtain the label that needs to be recognized for the challenge"""
        raise NotImplementedError

    def tactical_retreat(self, **kwargs) -> str | None:
        """skip unchoreographed challenges"""
        if self._label_alias.get(self._label):
            return self.status.CHALLENGE_CONTINUE

        q = quote(self._label, "utf8")
        logger.warning(
            "Types of challenges not yet scheduled",
            label=self._label,
            prompt=self._prompt,
            issue=f"https://github.com/QIN2DIM/hcaptcha-challenger/issues?q={q}",
        )
        return self.status.CHALLENGE_BACKCALL

    def mark_samples(self, ctx, *args, **kwargs):
        """Get the download link and locator of each challenge image"""
        raise NotImplementedError

    def download_images(self):
        prefix = ""
        if self._label:
            prefix = f"{time.time()}_{self._label_alias.get(self._label, self._label)}"
        runtime_dir = self.challenge_dir.joinpath(prefix)
        runtime_dir.mkdir(mode=0o777, parents=True, exist_ok=True)

        # Initialize the data container
        container = []
        for alias_, url_ in self._alias2url.items():
            challenge_img_path = runtime_dir.joinpath(f"{alias_}.png")
            self._alias2path.update({alias_: challenge_img_path})
            container.append((challenge_img_path, url_))

        return container

    def challenge(self, ctx, model, *args, **kwargs):
        """
        图像分类，元素点击，答案提交

        ### 性能瓶颈

        此部分图像分类基于 CPU 运行。如果服务器资源极其紧张，图像分类任务可能无法按时完成。
        根据实验结论来看，如果运行时内存少于 512MB，且仅有一个逻辑线程的话，基本上是与深度学习无缘了。

        ### 优雅永不过时

        `hCaptcha` 的挑战难度与 `reCaptcha v2` 不在一个级别。
        这里只要正确率上去就行，也即正确图片覆盖更多，通过率越高（即使因此多点了几个干扰项也无妨）。
        所以这里要将置信度尽可能地调低（未经针对训练的模型本来就是用来猜的）。

        :return:
        """
        raise NotImplementedError

    def is_success(self, ctx, *args, **kwargs):
        """
        判断挑战是否成功的复杂逻辑

        # 首轮测试后判断短时间内页内是否存在可点击的拼图元素
        # hcaptcha 最多两轮验证，一般情况下，账号信息有误仅会执行一轮，然后返回登录窗格提示密码错误
        # 其次是被识别为自动化控制，这种情况也是仅执行一轮，回到登录窗格提示“返回数据错误”

        经过首轮识别点击后，出现四种结果:
            1. 直接通过验证（小概率）
            2. 进入第二轮（正常情况）
                通过短时间内可否继续点击拼图来断言是否陷入第二轮测试
            3. 要求重试（小概率）
                特征被识别|网络波动|被标记的（代理）IP
            4. 通过验证，弹出 2FA 双重认证
              无法处理，任务结束

        :param ctx: 挑战者驱动上下文
        :return:
        """
        raise NotImplementedError

    def anti_checkbox(self, ctx, *args, **kwargs):
        raise NotImplementedError

    def anti_hcaptcha(self, ctx, *args, **kwargs) -> bool | str:
        """
        Handle hcaptcha challenge

        ## Method

        具体思路是：
        1. 进入 hcaptcha iframe
        2. 获取图像标签
            需要加入判断，有时候 `hcaptcha` 计算的威胁程度极低，会直接让你过，
            于是图像标签之类的元素都不会加载在网页上。
        3. 获取各个挑战图片的下载链接及网页元素位置
        4. 图片下载，分类
            需要用一些技术手段缩短这部分操作的耗时。人机挑战有时间限制。
        5. 对正确的图片进行点击
        6. 提交答案
        7. 判断挑战是否成功
            一般情况下 `hcaptcha` 的验证有两轮，
            而 `recaptcha vc2` 之类的人机挑战就说不准了，可能程序一晚上都在“循环”。

        ## Reference

        M. I. Hossen and X. Hei, "A Low-Cost Attack against the hCaptcha System," 2021 IEEE Security
        and Privacy Workshops (SPW), 2021, pp. 422-431, doi: 10.1109/SPW53761.2021.00061.

        > ps:该篇文章中的部分内容已过时，如今的 hcaptcha challenge 远没有作者说的那么容易应付。
        :param ctx:
        :return:
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Time       : 2023/8/19 17:52
# Author     : QIN2DIM
# GitHub     : https://github.com/QIN2DIM
# Description:
from __future__ import annotations

import asyncio
import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Tuple, Iterable

from httpx import AsyncClient


@dataclass
class AshFramework(ABC):
    container: Iterable

    @classmethod
    def from_container(cls, container):
        if sys.platform.startswith("win") or "cygwin" in sys.platform:
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        else:
            asyncio.set_event_loop(asyncio.new_event_loop())
        return cls(container=container)

    @abstractmethod
    async def control_driver(self, context: Any, client: AsyncClient):
        raise NotImplementedError

    async def subvert(self):
        if not self.container:
            return
        async with AsyncClient() as client:
            task_list = [self.control_driver(context, client) for context in self.container]
            await asyncio.gather(*task_list)

    def execute(self):
        asyncio.run(self.subvert())


@dataclass
class ImageDownloader(AshFramework):
    async def control_driver(self, context: Any, client: AsyncClient):
        (img_path, url) = context
        resp = await client.get(url)
        img_path.write_bytes(resp.content)


def download_images(container: Iterable[Tuple[Path, str]]):
    """
    Download Challenge Image

    ### hcaptcha has a challenge duration limit

    If the page element is not manipulated for a period of time,
    the <iframe> box will disappear and the previously acquired Element Locator will be out of date.
    Need to use some modern methods to shorten the time of `getting the dataset` as much as possible.

    ### Solution

    1. Coroutine Downloader
      Use the coroutine-based method to _pull the image to the local, the best practice (this method).
      In the case of poor network, _pull efficiency is at least 10 times faster than traversal download.

    2. Screen cut
      There is some difficulty in coding.
      Directly intercept nine pictures of the target area, and use the tool function to cut and identify them.
      Need to weave the locator index yourself.

    :param container:
    :return:
    """
    ImageDownloader(container).execute()



# -*- coding: utf-8 -*-
# Time       : 2023/8/25 14:00
# Author     : QIN2DIM
# GitHub     : https://github.com/QIN2DIM
# Description:
from __future__ import annotations

import random
import re
import time
import warnings
from dataclasses import dataclass

from loguru import logger
from selenium.common.exceptions import (
    WebDriverException,
    TimeoutException,
    NoSuchElementException,
    ElementNotInteractableException,
    ElementNotVisibleException,
    StaleElementReferenceException,
    ElementClickInterceptedException,
    InvalidArgumentException,
)
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait
from undetected_chromedriver import Chrome

from hcaptcha_challenger.agents.exceptions import ChallengePassed, LabelNotFoundException
from hcaptcha_challenger.agents.skeleton import Skeleton
from hcaptcha_challenger.components.image_downloader import download_images
from hcaptcha_challenger.components.prompt_handler import split_prompt_message, label_cleaning

warnings.filterwarnings("ignore", category=DeprecationWarning)


@dataclass
class SeleniumAgent(Skeleton):
    _threat = 0

    def switch_to_challenge_frame(self, ctx, **kwargs):
        WebDriverWait(ctx, 15, ignored_exceptions=(ElementNotVisibleException,)).until(
            EC.frame_to_be_available_and_switch_to_it(
                (By.XPATH, "//iframe[contains(@src,'#frame=challenge')]")
            )
        )

    def get_label(self, ctx, **kwargs):
        # Scan and determine the type of challenge.
        for _ in range(3):
            try:
                label_obj = WebDriverWait(
                    ctx, 5, ignored_exceptions=(ElementNotVisibleException,)
                ).until(EC.presence_of_element_located((By.XPATH, "//h2[@class='prompt-text']")))
            except TimeoutException:
                raise ChallengePassed("Man-machine challenge unexpectedly passed")
            else:
                self._prompt = label_obj.text
                if self._prompt:
                    break
                time.sleep(1)
                continue
        # Skip the `draw challenge`
        else:
            logger.debug(
                "Pass challenge", challenge="image_label_area_select", site_link=ctx.current_url
            )
            return self.status.CHALLENGE_BACKCALL

        # Continue the `click challenge`
        try:
            _label = split_prompt_message(prompt_message=self._prompt, lang=self.modelhub.lang)
        except (AttributeError, IndexError):
            raise LabelNotFoundException("Get the exception label object")
        else:
            self._label = label_cleaning(_label)

    def mark_samples(self, ctx, *args, **kwargs):
        # 等待图片加载完成
        try:
            WebDriverWait(ctx, 5, ignored_exceptions=(ElementNotVisibleException,)).until(
                EC.presence_of_all_elements_located((By.XPATH, "//div[@class='task-image']"))
            )
        except TimeoutException:
            try:
                ctx.switch_to.default_content()
                WebDriverWait(ctx, 1, 0.1).until(
                    EC.visibility_of_element_located(
                        (By.XPATH, "//div[contains(@class,'hcaptcha-success')]")
                    )
                )
                return self.status.CHALLENGE_SUCCESS
            except WebDriverException:
                return self.status.CHALLENGE_BACKCALL

        time.sleep(0.3)

        # DOM 定位元素
        samples = ctx.find_elements(By.XPATH, "//div[@class='task-image']")
        for sample in samples:
            alias = sample.get_attribute("aria-label")
            while True:
                try:
                    image_style = sample.find_element(By.CLASS_NAME, "image").get_attribute("style")
                    url = re.split(r'[(")]', image_style)[2]
                    self._alias2url.update({alias: url})
                    break
                except IndexError:
                    continue
            self._alias2locator.update({alias: sample})

    def download_images(self):
        container = super().download_images()
        download_images(container=container)

    def challenge(self, ctx, model, *args, **kwargs):
        ta = []
        # {{< IMAGE CLASSIFICATION >}}
        for alias, path in self._alias2path.items():
            t0 = time.time()
            result = model.execute(img_stream=path.read_bytes())
            ta.append(time.time() - t0)
            # Pass: Hit at least one object
            if result:
                try:
                    time.sleep(random.uniform(0.2, 0.3))
                    self._alias2locator[alias].click()
                except StaleElementReferenceException:
                    pass
                except WebDriverException as err:
                    logger.warning(err)

        # {{< SUBMIT ANSWER >}}
        try:
            WebDriverWait(ctx, 15, ignored_exceptions=(ElementClickInterceptedException,)).until(
                EC.element_to_be_clickable((By.XPATH, "//div[@class='button-submit button']"))
            ).click()
        except ElementClickInterceptedException:
            pass
        except WebDriverException as err:
            logger.exception(err)

    def is_success(self, ctx, *args, **kwargs):
        def is_challenge_image_clickable():
            try:
                WebDriverWait(ctx, 1, poll_frequency=0.1).until(
                    EC.presence_of_element_located((By.XPATH, "//div[@class='task-image']"))
                )
                return True
            except TimeoutException:
                return False

        def is_flagged_flow():
            try:
                WebDriverWait(ctx, 1.2, poll_frequency=0.1).until(
                    EC.visibility_of_element_located((By.XPATH, "//div[@class='error-text']"))
                )
                self._threat += 1
                if self._threat > 3:
                    logger.warning("Your proxy IP may have been flagged")
                return True
            except TimeoutException:
                return False

        time.sleep(1)
        if is_flagged_flow():
            return self.status.CHALLENGE_RETRY
        if is_challenge_image_clickable():
            return self.status.CHALLENGE_CONTINUE
        return self.status.CHALLENGE_SUCCESS

    def anti_checkbox(self, ctx, *args, **kwargs):
        for _ in range(8):
            try:
                # [👻] 进入复选框
                WebDriverWait(ctx, 2, ignored_exceptions=(ElementNotVisibleException,)).until(
                    EC.frame_to_be_available_and_switch_to_it(
                        (By.XPATH, "//iframe[contains(@title,'checkbox')]")
                    )
                )
                # [👻] 点击复选框
                WebDriverWait(ctx, 2).until(EC.element_to_be_clickable((By.ID, "checkbox"))).click()
                return True
            except (TimeoutException, InvalidArgumentException):
                pass
            finally:
                # [👻] 回到主线剧情
                ctx.switch_to.default_content()

    def anti_hcaptcha(self, ctx: Chrome, *args, **kwargs) -> bool | str:
        # [👻] 它來了！
        try:
            # If it cycles more than twice, your IP has been blacklisted
            for index in range(3):
                # [👻] 進入挑戰框架
                self.switch_to_challenge_frame(ctx)

                # [👻] 獲取挑戰標簽
                result = self.get_label(ctx)
                if result == self.status.CHALLENGE_BACKCALL:
                    ctx.switch_to.default_content()
                    return result

                # [👻] 編排定位器索引
                result = self.mark_samples(ctx)
                if result in [self.status.CHALLENGE_SUCCESS, self.status.CHALLENGE_CONTINUE]:
                    ctx.switch_to.default_content()
                    return result

                # [👻] 拉取挑戰圖片
                self.download_images()

                # [👻] 滤除无法处理的挑战类别
                result = self.tactical_retreat()
                if result == self.status.CHALLENGE_BACKCALL:
                    ctx.switch_to.default_content()
                    return result

                # [👻] 注册解决方案
                # 根据挑战类型自动匹配不同的模型
                model = self._match_solution()

                # [👻] 識別|點擊|提交
                self.challenge(ctx, model)

                # [👻] 輪詢控制臺響應
                result = self.is_success(ctx)

                ctx.switch_to.default_content()
                if result in [
                    self.status.CHALLENGE_SUCCESS,
                    self.status.CHALLENGE_CRASH,
                    self.status.CHALLENGE_RETRY,
                ]:
                    return result

        except WebDriverException as err:
            logger.exception(err)
            ctx.switch_to.default_content()
            return self.status.CHALLENGE_CRASH


class ArmorUtils:
    @staticmethod
    def face_the_checkbox(ctx: Chrome) -> bool | None:
        try:
            WebDriverWait(ctx, 8, ignored_exceptions=(WebDriverException,)).until(
                EC.presence_of_element_located((By.XPATH, "//iframe[contains(@title,'checkbox')]"))
            )
            return True
        except TimeoutException:
            return False

    @staticmethod
    def get_hcaptcha_response(ctx: Chrome) -> str | None:
        return ctx.execute_script("hcaptcha.getResponse()")

    @staticmethod
    def refresh(ctx: Chrome) -> bool | None:
        try:
            ctx.find_element(By.XPATH, "//div[@class='refresh button']").click()
        except (NoSuchElementException, ElementNotInteractableException):
            return False
        return True
